/*
 Stephan Kottler
 kottlers@informatik.uni-tuebingen.de
 *
 */

#include "Statistics.hpp"
#include "BasicSolver.hpp"
#include "ConstraintDB.hpp"
#include "Types.hpp"
#include "Settings.hpp"
#include "IFStream.hpp"
#include "Settings.hpp"
#include "Random.hpp"
//#include "Flag.hpp"
#include "BitVector.hpp"
#include "CatActHeap.hpp"
#include <locale.h>


Statistics::Statistics(Settings& sets):
    settings                        (sets),
    nbr_conflicts                   ( 0 ),
    nbr_propagations                ( 0 ),
    nbr_act_decisions               ( 0 ),
    nbr_rnd_decisions               ( 0 ),
    nbr_restarts                    ( 0 ),
    nbr_local_jumpbacks				( 0 ),
    max_backjumping_distance        ( 0 ),
    sum_backjumping                 ( 0 ),
    implicit_card_binaries			( 0 ),
    //nbr_propagate_shortcut_failed   ( 0 ),
    //nbr_propagate_shortcut_worked   ( 0 ),
    //nbr_search_steps_for_watcher    ( 0 ),
    nbr_expensive_cls_bmp           ( 0 ),
    nbr_expensive_var_bmp           ( 0 ),
    nbr_flipped_prop_learnts		(0),
    nbr_learnt_reductions           ( 0 ),
    total_lemma_shortening          ( 0 ),
    nbr_bin_subsumes_tern           ( 0 ),
    nbr_simplify_execs              ( 0 ),
    nbr_removed_cls_lits            ( 0 ),
    nbr_simplify_worked             ( 0 ),
    nbr_subsumed_worked				( 0 ),
    nbr_subsumed_shrinked			( 0 ),
    nbr_subsumed_shrink_card		( 0 ),
    nbr_cat_minimize_lits			( 0 ),
    nbr_cat_minimize_worked			( 0 ),
    nbr_cat_minimize_fail_redundant	( 0 ),
    nbr_bigsimplify_shrinked        ( 0 ),
    max_dec_level                  	( 0 ),
    max_assigned					( 0 ),
    nbr_big_simp_card_shrink        ( 0 ),
    nbr_closure_create_bin			( 0 ),
    nbr_closure_create_tern			( 0 ),
    nbr_expensive_cls_stamp			( 0 ),
    nbr_panic_gc					( 0 ),
    vertex_count					( 0 ),
    edge_count					 	( 0 ),
    nbr_arc_const_check_worked		( 0 ),
    nbr_arc_const_check_failed		( 0 ),
    nbr_arc_const_levels_omit		( 0 ),
    parse_timestmp                  ( 0.0 ),
    finish_timestmp                 ( 0.0 )
{}


void Statistics::storeBackjumpingDistance(const int dist){
    /*uint32_t index = dist / 50;
    if(index >= backjumping_distances.size()){    backjumping_distances.resize(index+1,0);}
    backjumping_distances[index] += 1;*/

    //nur alternative
    sum_backjumping += dist;
    set_max_int(max_backjumping_distance,dist);
}


double Statistics::cpuSystemTime()
{
    struct rusage ru;
    getrusage(RUSAGE_SELF, &ru);
    return (double)ru.ru_stime.tv_sec + (double)ru.ru_stime.tv_usec / 1000000.0;
};

double Statistics::cpuUserTime()
{
    struct rusage ru;
    getrusage(RUSAGE_SELF, &ru);
    return (double)ru.ru_utime.tv_sec + (double)ru.ru_utime.tv_usec / 1000000.0;
};


uint64_t Statistics::usedMemoryBytes()
{
    char    path[256];
    sprintf(path, "/proc/%d/statm", getpid());
    FILE*   in = fopen(path, "rb");
    if (in == NULL) return 0;
    uint64_t     value;
    fscanf(in, "%lld", &value);
    fclose(in);
    return value * (uint64_t)getpagesize();
};

void Statistics::printStatsProgressHead(OFStream* out){
	assert(out != NULL);
	out->writeString("|       Starts     |  Decisions  | Propagations |  Conflicts  |        Literals        |   DB Red");
}

void Statistics::printStatsProgress(OFStream* out, uint64_t literals){
	double  used_bytes = (double)Statistics::usedMemoryBytes();
	assert(out != NULL);
	out->writeString("| %7u %8.1fs| %12llu| %13llu| %12llu| %11llu  %6.2f[MB]| %8u",
            nbr_restarts, cpuUserTime(),
            nbr_rnd_decisions+nbr_act_decisions,
            nbr_propagations,
            nbr_conflicts,
            literals, (used_bytes / (double)MEGA),
            nbr_learnt_reductions);
}


void Statistics::printFinalStatistics(OFStream& out, const ConstraintDB* sp){

    finish_timestmp                = cpuUserTime();
    double  cpu_s_time(Statistics::cpuSystemTime());
    double  used_bytes = (double)Statistics::usedMemoryBytes();
    out.writeString("| ");
    out.writeString("|Statistics: ");
    if(sp != NULL){
    out.writeString("|Variables:                   %u "             ,sp->numberVariables());
    out.writeString("|Total Clauses:               %u "             ,sp->numberAllClauses());
    out.writeString("|Learnt Clauses:              %u "             ,sp->numberLearntLongClses());
    out.writeString("|Binary Clauses:              %u "             ,sp->numberBinClses());
    out.writeString("|Ternary Clauses:             %u "             ,sp->numberTernClses());
    out.writeString("|Short Clauses:               %u "             ,sp->numberShortClses());
    out.writeString("|CNF Clause Literals:         %llu "           ,sp->numberLiteralsInLongFixClses()+sp->numberLiteralsInShortClses());
    out.writeString("|Learnt Literals:             %llu "           ,sp->numberLiteralsInLongLearntClses());
    }

    out.writeString("|CPU User Time:               %g [sec]"        ,finish_timestmp);
    out.writeString("|CPU System Time:             %g [sec]"        ,cpu_s_time);
    out.writeString("|Parse Time:                  %g [sec]"        ,parse_timestmp);
    out.writeString("|Used Memory:                 %.2f [MB] "      ,used_bytes / (double)MEGA);

    out.writeString("|Starts:                      %u "             ,nbr_restarts);
    //out.writeString("|Local Backjumps:             %u "             ,nbr_local_jumpbacks);
    out.writeString("|Random Decisions:            %llu (%.2f %%)"  ,nbr_rnd_decisions,(100.0*(double)nbr_rnd_decisions)/(double)(nbr_rnd_decisions + nbr_act_decisions));
    out.writeString("|Activity Decisions:          %llu (%.2f %%)"  ,nbr_act_decisions,(100.0*(double)nbr_act_decisions)/(double)(nbr_rnd_decisions + nbr_act_decisions));
    out.writeString("|Conflicts:                   %llu "           ,nbr_conflicts);
    out.writeString("|Propagations:                %llu "           ,nbr_propagations);
    out.writeString("|Propagations per Decision    %.2f "           ,(double)nbr_propagations/((double)(nbr_rnd_decisions+nbr_act_decisions)));

    /*out.writeString("|Expensive Variable Decay:    %u "             ,nbr_expensive_var_bmp);
    out.writeString("|Flipped Props worked:        %u "             ,nbr_flipped_prop_learnts);

    out.writeString("|Expensive Clause Decay:      %u "             ,nbr_expensive_cls_bmp);
    out.writeString("|Expensive Clause Stamp:      %u "             ,nbr_expensive_cls_stamp);
    out.writeString("|Simplify of clause DB:       %u "             ,nbr_simplify_execs);
    out.writeString("|Simplify worked:             %u "             ,nbr_simplify_worked);
    out.writeString("|Subsumption worked:          %u "             ,nbr_subsumed_worked);
    out.writeString("|Subsumption shrinked:        %u "             ,nbr_subsumed_shrinked);
    out.writeString("|Subsumption shrinked card:   %u "             ,nbr_subsumed_shrink_card);

    out.writeString("|Cat Minimize lits:           %u "             ,nbr_cat_minimize_lits);
    out.writeString("|Cat Minimize worked:         %u "             ,nbr_cat_minimize_worked);
    out.writeString("|Cat Minimize redundant fail: %u "             ,nbr_cat_minimize_fail_redundant);
    out.writeString("|Implicit Binaries by cards:  %llu "           ,implicit_card_binaries);

    out.writeString("|Big Simplify shrinked:       %u "             ,nbr_bigsimplify_shrinked);
    out.writeString("|Simplify reduced cls lit:    %u "             ,nbr_removed_cls_lits);
    out.writeString("|Reductions of clause DB:     %u "             ,nbr_learnt_reductions);
    out.writeString("|Total lemma shortening:      %llu "           ,total_lemma_shortening);
    out.writeString("|Binary subsumed ternary:     %llu "           ,nbr_bin_subsumes_tern);
    out.writeString("|Max Decision Level:          %u "             ,max_dec_level);
    out.writeString("|Max Assignments:             %u "             ,max_assigned);
    out.writeString("|Big Simp Card Shrink:        %u "             ,nbr_big_simp_card_shrink);

    out.writeString("|Closure create Bin:          %u "             ,nbr_closure_create_bin);
    out.writeString("|Closure create Tern:         %u "             ,nbr_closure_create_tern);

    out.writeString("|Arc const check invane:      %u "             ,nbr_arc_const_check_failed);
    out.writeString("|Arc const worked:            %u "             ,nbr_arc_const_check_worked);
    out.writeString("|Arc const levels omitted:    %u "             ,nbr_arc_const_levels_omit);
    */

    out.writeString("|Max Backjumping:             %d"              ,max_backjumping_distance);
    out.writeString("|Sum Backjumping:             %llu"            ,sum_backjumping);
    out.writeString("|Backjumps per Conflict:      %.2f "           ,(double)sum_backjumping/(double)nbr_conflicts);



    out.writeString("| ");
    //printCVSVersionNumbers(out,"|");
}


static inline const char* polarityMode2String(PolarityMode p){
    switch(p){
        case Polarity_False:            return "Polarity False";
        case Polarity_True:             return "Polarity True";
        case Polarity_Rand:             return "Polarity Random";
        case Polarity_Prev:             return "Polarity Previous";
        case Polarity_Card:             return "Polarity Card";
    }
    assert(false);
    return "Polarity Unknown";
};


inline static std::string toString(double d){
    static char buff[15];
    sprintf(buff," %*.3f ",12,d);
    return std::string(buff);
}

inline static std::string toString(uint64_t l){
    static char buff[15];
    sprintf(buff," %12llu ",l);
    return std::string(buff);
}

inline static std::string toString(uint32_t i){
    static char buff[15];
    sprintf(buff," %12u ",i);
    return std::string(buff);
}

inline static std::string toString(int32_t i){
    static char buff[15];
    sprintf(buff," %12d ",i);
    return std::string(buff);
}

void Statistics::printFile(OFStream& out,const char* sep, const ConstraintDB* sp, char* result){

    finish_timestmp = Statistics::cpuUserTime();
    double  cpu_s_time(Statistics::cpuSystemTime());
    double  used_bytes = (double)Statistics::usedMemoryBytes();

    out.writeString("# Invocation: %s \n",settings.command_line.c_str());

    out.writeString("# ");
    out.writeString("rand dec freq: %d/1000, "      , settings.random_var_freq_permil);
    out.writeString("initial restart limit: %d, "   , settings.conflict_max_init);
    if(settings.timeout < 0){
    out.writeString("timeout:  no restriction, ");
    }else{
    out.writeString("timeout: %d sec, "             , settings.timeout);}
    out.writeString("pol mode:  %s, "               , polarityMode2String(settings.polarity_mode));
    out.writeString("\n ");

    out.writeString("#");
    std::string values  = "";

    if(sp != NULL){
    out.writeString("Variables     %s",sep);  values  += toString(sp->numberVariables()    ) + sep;
    out.writeString("Total Cls     %s",sep);  values  += toString(sp->numberAllClauses() ) 	 + sep;
    out.writeString("Learnt Cls    %s",sep);  values  += toString(sp->numberLearntLongClses()) + sep;
    out.writeString("Binary Cls    %s",sep);  values  += toString(sp->numberShortClses()   )   + sep;
    out.writeString("CNF Literals  %s",sep);  values  += toString(sp->numberLiteralsInLongFixClses()
																+ sp->numberLiteralsInShortClses()) + sep;
    out.writeString("Learnt Lits   %s",sep);  values  += toString(sp->numberLiteralsInLongLearntClses()) + sep;
    }

    out.writeString("Vertex Count  %s",sep);  values  += toString(vertex_count               ) + sep;
    out.writeString("Edge Count    %s",sep);  values  += toString(edge_count                 ) + sep;
    out.writeString("Grid Width    %s",sep);  values  += toString(settings.grid_width        ) + sep;
    out.writeString("Grid Height   %s",sep);  values  += toString(settings.grid_height       ) + sep;

    out.writeString("User Time     %s",sep);  values  += toString(finish_timestmp            ) + sep;
    out.writeString("System Time   %s",sep);  values  += toString(cpu_s_time                 ) + sep;
    out.writeString("Parse Time    %s",sep);  values  += toString(parse_timestmp             ) + sep;
    out.writeString("Memory        %s",sep);  values  += toString(used_bytes / (double)MEGA  ) + sep;

    out.writeString("Starts        %s",sep);  values  += toString(nbr_restarts               ) + sep;
    out.writeString("Local BJumps  %s",sep);  values  += toString(nbr_local_jumpbacks        ) + sep;
    out.writeString("Random Decs   %s",sep);  values  += toString(nbr_rnd_decisions          ) + sep;
    out.writeString("Activity Decs %s",sep);  values  += toString(nbr_act_decisions          ) + sep;
    out.writeString("Conflicts     %s",sep);  values  += toString(nbr_conflicts              ) + sep;
    out.writeString("Propagations  %s",sep);  values  += toString(nbr_propagations           ) + sep;

    out.writeString("Exp Var Decay %s",sep);  values  += toString(nbr_expensive_var_bmp      ) + sep;
    out.writeString("Fl Props work %s",sep);  values  += toString(nbr_flipped_prop_learnts   ) + sep;

    out.writeString("Exp Cls Decay %s",sep);  values  += toString(nbr_expensive_cls_bmp      ) + sep;
    out.writeString("Exp Cls Stamp %s",sep);  values  += toString(nbr_expensive_cls_stamp    ) + sep;
    out.writeString("Simp cls db   %s",sep);  values  += toString(nbr_simplify_execs         ) + sep;
    out.writeString("Simp del cls  %s",sep);  values  += toString(nbr_simplify_worked        ) + sep;
    out.writeString("BigSimp Shrin %s",sep);  values  += toString(nbr_bigsimplify_shrinked   ) + sep;
    out.writeString("Panic GC      %s",sep);  values  += toString(nbr_panic_gc  			 ) + sep;
    out.writeString("Simp del lits %s",sep);  values  += toString(nbr_removed_cls_lits       ) + sep;
    out.writeString("Reduce cdb    %s",sep);  values  += toString(nbr_learnt_reductions      ) + sep;
    out.writeString("Lemma shrink  %s",sep);  values  += toString(total_lemma_shortening     ) + sep;

    //out.writeString("Prop wt good  %s",sep);  values  += toString(nbr_propagate_shortcut_failed)+sep;
    //out.writeString("Prop wt bad   %s",sep);  values  += toString(nbr_propagate_shortcut_worked)+sep;
    //out.writeString("Steps Watcher %s",sep);  values  += toString(nbr_search_steps_for_watcher)+ sep;
    out.writeString("Max Dec Level %s",sep);  values  += toString(max_dec_level             ) + sep;
    out.writeString("Pro Cls Dec   %s",sep);  values  += toString(nbr_big_simp_card_shrink     ) + sep;
    out.writeString("Max Backjmp   %s",sep);  values  += toString(max_backjumping_distance   ) + sep;
    out.writeString("Sum Backjmp   %s",sep);  values  += toString(sum_backjumping            ) + sep;

    time_t rawtime; time(&rawtime); struct tm* timeinfo = localtime(&rawtime);
    char* date_time = asctime(timeinfo); *(date_time + strlen(date_time) - 1) = ' ';

    out.writeString("Date          %s",sep);  values  += " \"" + std::string(date_time) + "\" "+ sep;
    out.writeString("Result        %s",sep);  values  += std::string(result                  ) + sep;
    out.writeString("Instance      %s",sep);  values  += " \"" + std::string(settings.tgf_file_name) + "\" " + sep;

    out.writeString(" \n");

    out.writeString("%s \n",values.c_str());

}





void    Statistics::printGnuPlot(OFStream& out,const ConstraintDB* sp, char* result){

    printFile(out," ",sp,result);
}


void Statistics::printCSV(OFStream& out,const ConstraintDB* sp, char* result){

    printFile(out,"&",sp,result);

    out.writeString("\n");
}


